/*--------------------------------------------------------------------------

   glvViewer class
   glvViewer.cpp v1.0 7/November/2003
   Copyright (C) 2003 J. Ribelles

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

  --------------------------------------------------------------------------*/

//using namespace std;

//#define FLTK

#ifdef FLTK
#include <FL/glut.H>
#include <GL/glu.h>
#include <FL/Fl_Window.H>
#else
#include <GL/glut.h>
#endif
#include <math.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <string.h>

#include "glvViewer.h"

#define GLV_NO_BUTTON_PRESSED             3
#define GLV_LEFT_BUTTON_HOLDED            4
#define GLV_RIGHT_BUTTON_HOLDED           5
#define GLV_GLUT_LEFT_BUTTON_AND_SHIFT    6
#define GLV_LEFT_BUTTON_AND_SHIFT_HOLDED  7
#define GLV_GLUT_LEFT_BUTTON_AND_CTRL     8
#define GLV_LEFT_BUTTON_AND_CTRL_HOLDED   9

glvViewer *glvViewerv;              
int idWin, idMenu;

int	glvProjection= glvPerspective;
GLenum  glvNormalize= GL_TRUE;
GLfloat glvAspectWH, glvAspectHW, glvFovy= 45.0f;
int     glvViewportWidth,   glvViewportHeight;     // viewport size

short  mouseButtonPressed= GLV_NO_BUTTON_PRESSED;  // what button is pressed??
unsigned short  oldMouseX= 0, oldMouseY= 0;    // previous mouse coordinates

float  glvAngleX, glvAngleY;                   // X and Y rotation angles
int    ctrlRotation= 0;                        // to rotate the model or not
          
int glvFrames=0, glvTime, glvTimebase=0;
time_t glvTime_t, glvTimebase_t=(time_t)0;

float glvMatrixRotation[16]=   {1.0f, 0.0, 0.0f, 0.0f,
                                0.0f, 1.0, 0.0f, 0.0f,
                                0.0f, 0.0, 1.0f, 0.0f,
                                0.0f, 0.0, 0.0f, 1.0f},
      matrixTranslation[3]= {0.0, 0.0f, 0.0f};

int glvViewerWindowWidth= 800, glvViewerWindowHeight= 600,// main window size
    glvViewerWindowPosX=   50,  glvViewerWindowPosY=   50;// main window position

GLenum	glvPolygonMode=    GL_FILL, glvFrontFace=  GL_CCW,	
        glvCullFace=       GL_BACK, glvShadeModel= GL_SMOOTH,	
        glvEnableCullFace= GL_TRUE;

GLclampf glvRed= 0.24f, glvGreen= 0.47f, glvBlue= 0.60f, glvAlpha= 1.0f,  // background color
         glvOwnColor[]={0.0f, 0.0f, 0.0f, 0.0f};                          // user background color   

char glvGameMode[]= "640x480:8@60";

int  currentColor= glvBlueColor;

GLenum  glvEnableLight0=        GL_TRUE,
        glvEnableLighting=      GL_TRUE,
        glvEnableColorMaterial= GL_TRUE;

GLfloat glvLight0Position[]= {0.0f, 0.0f, 10.0f, 1.0f},
        glvLight0Diffuse[]=  {0.7f, 0.7f,  0.7f, 1.0f},
        glvLight0Specular[]= {0.2f, 0.2f,  0.2f, 1.0f},
        glvLight0Ambient[]=  {0.1f, 0.1f,  0.1f, 1.0f},
        glvModelAmbient[]=   {0.2f, 0.2f,  0.2f, 1.0f};

GLboolean glvLightTwoSide=     GL_TRUE,
          glvLightLocalViewer= GL_TRUE;

GLfloat glvMaterialDiffuse[]=   {0.7f, 0.7f, 0.7f, 1.0f},
        glvMaterialSpecular[]=  {0.2f, 0.2f, 0.2f, 1.0f},
        glvMaterialShininess[]= {50.0f},
        glvMaterialEmission[]=  {0.0f, 0.0f, 0.0f, 0.0f},
        glvMaterialAmbient[]=   {0.1f, 0.1f, 0.1f, 1.0f};

#ifdef FLTK
Fl_Window *fltkWin;
#endif


/*----------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
                       GLUT CALLBACK FUNCTIONS (EVENTS)
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
----------------------------------------------------------------------------*/



/** Set the callback functions for display, keyboard, reshape, mouse, motion
 *  and idle.
 */
void glvViewer::glvEvents (void) {
		
  glutDisplayFunc  (_glvDisplay);
  glutKeyboardFunc (_glvKeyboard);
  glutReshapeFunc  (_glvReshape);
  glutMouseFunc    (_glvMouse);
  glutMotionFunc   (_glvMotion);
  glutIdleFunc     (_glvIdle);
}


/** Display callback. 
 */
void glvViewer::_glvDisplay (void) {

  glClearColor (glvRed, glvGreen, glvBlue, glvAlpha);
  glClear (GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  
  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  glTranslatef (matrixTranslation[0]/100000.0, 
		matrixTranslation[1]/100000.0, 0.0f);
  glTranslatef (0,0,-1.5);
  glMultMatrixf (glvMatrixRotation);
  
  glvViewerv->glvMyDisplay();

#ifndef FLTK  
  glutSwapBuffers ();
#endif
}


/** Reshape callback.
 */
void glvViewer::_glvReshape (int width, int height) {
  
  glViewport (0, 0, width, height);
  glvViewportWidth=  width;
  glvViewportHeight= height;
  glvAspectWH= (GLfloat) width/height;
  glvAspectHW= (GLfloat) height/width;
  glvViewerv->glvSetProjection(glvProjection);
}


/** Keyboard callback. 
*/
void glvViewer::_glvKeyboard (unsigned char key, int px, int py) {

  switch (key) {
  case  27: exit (0); break;
  case 'q': exit (0); break;
  case 'Q': exit (0); break;
#ifndef FLTK  
    case 'l': glutLeaveGameMode();break;
#endif
  }
  glvViewerv->glvMyKeyboard(key, px, py);
}


/** Mouse callback.
 */
void glvViewer::_glvMouse (int button, int state, int x, int y) {
 
  switch (button) {
  case GLUT_LEFT_BUTTON: {
    if (state== GLUT_DOWN)
#ifndef FLTK
      if (glutGetModifiers()== GLUT_ACTIVE_SHIFT)
#else
      if (Fl::event_shift())
#endif
	mouseButtonPressed= GLV_GLUT_LEFT_BUTTON_AND_SHIFT;
      else 	      
#ifndef FLTK
	if (glutGetModifiers()== GLUT_ACTIVE_CTRL)
#else
        if (Fl::event_ctrl())
#endif
	  mouseButtonPressed= GLV_GLUT_LEFT_BUTTON_AND_CTRL;
	else {
	  mouseButtonPressed= GLUT_LEFT_BUTTON;
	  ctrlRotation= 0;
	}
    else {
      mouseButtonPressed= GLV_NO_BUTTON_PRESSED;
      if (ctrlRotation) { glvAngleX= 0; glvAngleY= 0;}
      //glvAngleX= (float)(x- oldMouseX)/ 5.0f;
      //glvAngleY= (float)(y- oldMouseY)/ 5.0f;
      glutPostRedisplay ();
    }
    break;
  }
  }
} 


/** Motion callback.
 */
void glvViewer::_glvMotion (int newMouseX, int newMouseY) {

  short int incMouseX, incMouseY;
  switch (mouseButtonPressed) {
    case GLUT_LEFT_BUTTON: {
      mouseButtonPressed= GLV_LEFT_BUTTON_HOLDED;
      oldMouseX= newMouseX; 
      oldMouseY= newMouseY;
      break;
    }
    case GLV_GLUT_LEFT_BUTTON_AND_SHIFT: {
      mouseButtonPressed= GLV_LEFT_BUTTON_AND_SHIFT_HOLDED;
      oldMouseX= newMouseX; 
      oldMouseY= newMouseY;
      break;
    }
    case GLV_GLUT_LEFT_BUTTON_AND_CTRL: {
      mouseButtonPressed= GLV_LEFT_BUTTON_AND_CTRL_HOLDED;
      oldMouseX= newMouseX; 
      oldMouseY= newMouseY;
      break;
    }
    case GLV_LEFT_BUTTON_HOLDED: {  // rotation
      glvAngleX= (float)(newMouseX- oldMouseX)/ 5.0f;
      glvAngleY= (float)(newMouseY- oldMouseY)/ 5.0f;
      glMatrixMode (GL_MODELVIEW);
      glLoadIdentity ();
      glRotatef (glvAngleX, 0.0f,1.0f,0.0f);
      glRotatef (glvAngleY, 1.0f,0.0f,0.0f);
      glMultMatrixf (glvMatrixRotation);
      glGetFloatv (GL_MODELVIEW_MATRIX, glvMatrixRotation);
      oldMouseX= newMouseX; 
      oldMouseY= newMouseY;
      ctrlRotation= 0;
      glutPostRedisplay ();
      break;
    }
    case GLV_LEFT_BUTTON_AND_CTRL_HOLDED: {  // zoom
      incMouseY= newMouseY- oldMouseY;
      glvFovy-= (incMouseY/ 10.0);
      if (glvFovy> 180.0f) glvFovy= 180.0f;
      else if (glvFovy< 1.0f) glvFovy= 1.0f;
      oldMouseX= newMouseX; 
      oldMouseY= newMouseY;
      glvViewerv->glvSetProjection(glvProjection);
      glutPostRedisplay ();
      break;
    }
    case GLV_LEFT_BUTTON_AND_SHIFT_HOLDED: { // translation
      incMouseX= newMouseX- oldMouseX;
      incMouseY= newMouseY- oldMouseY;
      matrixTranslation[0]+= incMouseX*100;
      matrixTranslation[1]-= incMouseY*100;
      oldMouseX= newMouseX; 
      oldMouseY= newMouseY;
      glutPostRedisplay ();
      break;
    }
  }
}


char cfps[8];
/** Idle callback, rotate the object and show the fps.
 */
void glvViewer::_glvIdle (void) {

  if ((mouseButtonPressed!= GLUT_LEFT_BUTTON) &&
      (mouseButtonPressed!= GLV_LEFT_BUTTON_HOLDED)) {
    glMatrixMode (GL_MODELVIEW);
    glLoadIdentity ();
    glRotatef (glvAngleX/10.0f, 0.0f,1.0f,0.0f);
    glRotatef (glvAngleY/10.0f, 1.0f,0.0f,0.0f);
    glMultMatrixf (glvMatrixRotation);
    glGetFloatv (GL_MODELVIEW_MATRIX, glvMatrixRotation);
    
    glvFrames++;
#ifndef FLTK
    glvTime= glutGet(GLUT_ELAPSED_TIME);
    if (glvTime - glvTimebase > 1000) {
      int fps = (int) (glvFrames*1000.0/(glvTime-glvTimebase));
      sprintf (cfps, "%d FPS", fps);
      if (!glutGameModeGet(GLUT_GAME_MODE_ACTIVE)) glutSetWindowTitle(cfps);
      glvTimebase= glvTime;		
      glvFrames= 0;
    }
#else
    glvTime_t= time(NULL);
    if (glvTime_t - glvTimebase_t > 0) {
      int fps = glvFrames;
      sprintf (cfps, "%d FPS", fps);
      fltkWin->label(cfps);
      glvTimebase_t= glvTime_t;
      glvFrames= 0;
    }
#endif
    glutPostRedisplay ();
  } else
    ctrlRotation= 1;
}



/*----------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
                         GLUT POPUP MENU FUNCTIONS
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
----------------------------------------------------------------------------*/



/** Attend the selected menu option.
 */
void glvViewer::_glvMenu (int option) {
  
#ifndef FLTK  
  switch (option) {
  case  0: glShadeModel (GL_SMOOTH); glvShadeModel= GL_SMOOTH; break;
  case  1: glShadeModel (GL_FLAT); glvShadeModel=  GL_FLAT;    break;
  case  2: glPolygonMode (GL_FRONT_AND_BACK, GL_FILL); glvPolygonMode= GL_FILL; break;
  case  3: glPolygonMode (GL_FRONT_AND_BACK, GL_LINE); glvPolygonMode= GL_LINE; break;
  case  4: glFrontFace (GL_CW);  glvFrontFace= GL_CW;  break;
  case  5: glFrontFace (GL_CCW); glvFrontFace= GL_CCW;  break;
  case  6: glCullFace (GL_FRONT); glvCullFace= GL_FRONT; break;
  case  7: glCullFace (GL_BACK);  glvCullFace= GL_BACK;  break;
  case  8: glEnable (GL_CULL_FACE);  glvEnableCullFace= GL_TRUE; break;
  case  9: glDisable (GL_CULL_FACE); glvEnableCullFace= GL_FALSE; break;
  case 10: currentColor= glvWhiteColor; 
           glvRed=   1.00f; 
           glvGreen= 1.00f; 
           glvBlue=  1.00f;                            break;
  case 11: currentColor= glvBlackColor; 
           glvRed=   0.00f; 
	   glvGreen= 0.00f; 
	   glvBlue=  0.00f;                            break;
  case 12: glvProjection= glvPerspective; 
           glvViewerv->glvSetProjection(glvProjection);break;
  case 13: glvProjection= glvParallel;    
           glvViewerv->glvSetProjection(glvProjection);break;
  case 14: currentColor= glvBlueColor; 
           glvRed=   0.24f; 
	   glvGreen= 0.47f; 
	   glvBlue=  0.60f;                            break;
  case 15: glEnable  (GL_LIGHT0);  glvEnableLight0= GL_TRUE;  break;
  case 16: glEnable  (GL_LIGHTING); glvEnableLighting= GL_TRUE;  break;
  case 17: glDisable (GL_LIGHT0); glvEnableLight0= GL_FALSE;   break;
  case 18: glDisable (GL_LIGHTING);  glvEnableLighting= GL_FALSE;  break;
  case 19: glEnable  (GL_COLOR_MATERIAL); glvEnableColorMaterial= GL_TRUE; break;
  case 20: glDisable (GL_COLOR_MATERIAL); glvEnableColorMaterial= GL_FALSE; break;
  case 21: currentColor= glvMyOwnColor; 
           glvRed=   glvOwnColor[0]; 
	   glvGreen= glvOwnColor[1]; 
	   glvBlue=  glvOwnColor[2]; 
	   glvAlpha= glvOwnColor[3];                   break;
  case 22: glvViewerv->glvSaveImage();                 break;
  case 50: glutReshapeWindow(350, 300);                break;
  case 51: glutReshapeWindow(720, 576);                break;
  case 52: glutReshapeWindow(800, 600);                break;
  case 53: glutReshapeWindow(1024, 768);               break;
  case 58: {
    glutGameModeString(glvGameMode);
    if (glutGameModeGet(GLUT_GAME_MODE_POSSIBLE)) {
      glutEnterGameMode();
      glvViewerv->glvEvents();	
      glvViewerv->glvRenderOptions();
      glvViewerv->glvMyRenderOptions();
    }
    break;}
  case 59: glutFullScreen();                           break;
  case 99: glvViewerv->glvSaveOptions();               break;
  case 100: glvViewerv->glvReloadOptions();            break;
  }
  glutPostRedisplay ();
#endif  
}


/** Make the glut popup menu. 
 */
void glvViewer::glvMenu (void) {
  
  idMenu= glutCreateMenu (_glvMenu);
  glutAttachMenu (GLUT_RIGHT_BUTTON);

#ifndef FLTK  
  int idLMenu= glutCreateMenu (_glvMenu);
  glutAddMenuEntry ("White", 10);
  glutAddMenuEntry ("Black", 11);
  glutAddMenuEntry ("Blue",  14);
  glutAddMenuEntry ("My Own Color",  21);
  glutSetMenu (idMenu);
  glutAddSubMenu ("Clear Color", idLMenu);

  idLMenu= glutCreateMenu (_glvMenu);
  glutAddMenuEntry ("350x300", 50);
  glutAddMenuEntry ("720x576 (Pal)", 51);
  glutAddMenuEntry ("800x600",  52);
  glutAddMenuEntry ("1024x768", 53);
  glutAddMenuEntry ("Game mode", 58);
  glutAddMenuEntry ("Full screen", 59);
  glutSetMenu (idMenu);
  glutAddSubMenu ("Window Size", idLMenu);

  idLMenu= glutCreateMenu (_glvMenu);
  glutAddMenuEntry ("Smooth", 0);
  glutAddMenuEntry ("Flat", 1);
  glutSetMenu (idMenu);
  glutAddSubMenu ("Shade Model", idLMenu);

  idLMenu= glutCreateMenu (_glvMenu);
  glutAddMenuEntry ("Fill", 2);
  glutAddMenuEntry ("Line", 3);
  glutSetMenu (idMenu);
  glutAddSubMenu ("Polygon Mode", idLMenu);

  idLMenu= glutCreateMenu (_glvMenu);
  glutAddMenuEntry ("CW", 4);
  glutAddMenuEntry ("CCW", 5);
  glutSetMenu (idMenu);
  glutAddSubMenu ("Front Face", idLMenu);

  idLMenu= glutCreateMenu (_glvMenu);
  glutAddMenuEntry ("Front", 6);
  glutAddMenuEntry ("Back", 7);
  glutSetMenu (idMenu);
  glutAddSubMenu ("Cull Face", idLMenu);

  idLMenu= glutCreateMenu (_glvMenu);
  glutAddMenuEntry ("Enable", 8);
  glutAddMenuEntry ("Disable", 9);
  glutSetMenu (idMenu);
  glutAddSubMenu ("Backface Cull", idLMenu);

  idLMenu= glutCreateMenu (_glvMenu);
  glutAddMenuEntry ("Perspective", 12);
  glutAddMenuEntry ("Parallel", 13);
  glutSetMenu (idMenu);
  glutAddSubMenu ("Projection", idLMenu);

  idLMenu= glutCreateMenu (_glvMenu);
  glutAddMenuEntry ("Enable", 16);
  glutAddMenuEntry ("Disable", 18);
  glutAddMenuEntry ("Enable  Light 0", 15);
  glutAddMenuEntry ("Disable Light 0", 17);
  glutAddMenuEntry ("Enable  Color material", 19);
  glutAddMenuEntry ("Disable Color material", 20);
  glutSetMenu (idMenu);
  glutAddSubMenu ("Lighting", idLMenu);

  glutAddMenuEntry ("Save image", 22);
  glutAddMenuEntry ("Save options", 99);
  glutAddMenuEntry ("Reload options", 100);
#endif  
  
  glvMyMenu(idMenu);
}



/*----------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
                         RENDER OPTIONS
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
----------------------------------------------------------------------------*/



/** Set all the render options.
 */
void glvViewer::glvRenderOptions (void) {
  
  if (glvNormalize) glEnable (GL_NORMALIZE);
  else glDisable (GL_NORMALIZE);

  glEnable        (GL_DEPTH_TEST);
  glPolygonMode   (GL_FRONT_AND_BACK, glvPolygonMode);
  glFrontFace     (glvFrontFace);
  glCullFace      (glvCullFace);
  glShadeModel    (glvShadeModel);
  glDepthFunc     (GL_LEQUAL);

  if (glvEnableCullFace) glEnable (GL_CULL_FACE);
  else glDisable (GL_CULL_FACE);
  glLightModelfv (GL_LIGHT_MODEL_AMBIENT,      glvModelAmbient);
  glLightModeli  (GL_LIGHT_MODEL_TWO_SIDE,     glvLightTwoSide);
  glLightModeli  (GL_LIGHT_MODEL_LOCAL_VIEWER, glvLightLocalViewer);	

    glMatrixMode (GL_MODELVIEW);
    glLoadIdentity();
    glLightfv (GL_LIGHT0, GL_POSITION, glvLight0Position);
    glMultMatrixf (glvMatrixRotation);

  glLightfv (GL_LIGHT0, GL_DIFFUSE,  glvLight0Diffuse);
  glLightfv (GL_LIGHT0, GL_SPECULAR, glvLight0Specular);
  glLightfv (GL_LIGHT0, GL_AMBIENT,  glvLight0Ambient);
  if (glvEnableLight0)   glEnable  (GL_LIGHT0);
  else glDisable (GL_LIGHT0);
  if (glvEnableLighting) glEnable  (GL_LIGHTING);
  else glDisable (GL_LIGHTING);

  glMaterialfv (GL_FRONT, GL_AMBIENT,   glvMaterialAmbient);
  glMaterialfv (GL_FRONT, GL_DIFFUSE,   glvMaterialDiffuse);
  glMaterialfv (GL_FRONT, GL_SPECULAR,  glvMaterialSpecular);
  glMaterialfv (GL_FRONT, GL_SHININESS, glvMaterialShininess);
  glMaterialfv (GL_FRONT, GL_EMISSION,  glvMaterialEmission);
  glColorMaterial (GL_FRONT, GL_DIFFUSE);
  if (glvEnableColorMaterial) glEnable (GL_COLOR_MATERIAL);
  else glDisable (GL_COLOR_MATERIAL);

  glvMyRenderOptions();
}



/*----------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
                         PUBLIC FUNCTIONS
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
----------------------------------------------------------------------------*/



/** Init OpenGL, show the window, and call to glvMenu(), glvEvents() 
 *  and glvRenderOptions().
 */
void glvViewer::glvInit (int argc, char **argv) {

  // init OpenGL and init the main window
  glutInit               (&argc, argv);
  glutInitWindowSize     (glvViewerWindowWidth, glvViewerWindowHeight);
  glutInitWindowPosition (glvViewerWindowPosX, glvViewerWindowPosY);
  glutInitDisplayMode    (GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH); 
  idWin= glutCreateWindow       ("OpenGL Viewer");

  glvMenu();
  glvEvents();	
  glvRenderOptions();
}


/** Init OpenGL when using FLTK, show the window at (7,33), 
 *  and call to glvMenu(), glvEvents() and glvRenderOptions().
 *  w: pointer to the Fl_Window
 */
void glvViewer::glvInitFLTK (void *w) {

  // init OpenGL and init the main window
  glutInitWindowSize     (glvViewerWindowWidth, glvViewerWindowHeight);
  glutInitWindowPosition (7, 33);
  glutInitDisplayMode    (GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH); 
  idWin= glutCreateWindow       ("OpenGL Viewer");
#ifdef FLTK
  fltkWin= (Fl_Window *)w;
#endif
  glvMenu();
  glvEvents();	
  glvRenderOptions();
}


/** Call to glutMainLoop().
 */
void glvViewer::glvLoop (void) {glutMainLoop();}


/** Constructor: Read the file of options: glv_options.
 */
glvViewer::glvViewer (void) {

  glvViewerv= this;
  ifstream in ("glv_options");
  
  if (in) {
    glvRead(in);
    in.close();
    }
}


/** Return the current projection.
 *  Result: glvParallel or glvPerspective.
 */
int glvViewer::glvGetProjection (void) {return glvProjection;}


/** Set the projection.
 * mode: glvParallel or glvPerspective.
 */
void glvViewer::glvSetProjection (int mode) {

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  
  glvProjection= mode;

  switch (glvProjection) {
  case glvParallel: {   
    if (glvAspectWH> 1) 
      glOrtho (-0.5f*glvAspectWH*glvFovy*0.03, 0.5f*glvAspectWH*glvFovy*0.03, 
	       -0.5f*glvFovy*0.03, 0.5f*glvFovy*0.03, 
	       -1.0f, 2.0f);
    else
      glOrtho (-0.5f*glvFovy*0.03, 0.5f*glvFovy*0.03, 
	       -0.5f*glvAspectHW*glvFovy*0.03, 0.5f*glvAspectHW*glvFovy*0.03, 
	       -1.0f, 2.0f);
    break;
  }
  case glvPerspective: {
    gluPerspective (glvFovy, glvAspectWH, .2f, 3.0f);
    break;
  }
  }
}


/** Set the fovy and call to  glvSetProjection().
 *  fovy: the angle of the gluPerspective function
 */
void glvViewer::glvSetFovy (float fovy) {

  glvFovy= fovy;
  glvSetProjection(glvProjection);
}


/** Set the clear color.
 *  color: glvBlackColor, glvWhiteColor, glvBlueColor or glvMyOwnColor
 */
void glvViewer::glvSetClearColor (int color) {

  switch (color) {
  case glvWhiteColor: 
    currentColor= glvWhiteColor; 
    glvRed=   1.00f; 
    glvGreen= 1.00f; 
    glvBlue=  1.00f;                            
    break;
  case glvBlackColor: 
    currentColor= glvBlackColor; 
    glvRed=   0.00f; 
    glvGreen= 0.00f; 
    glvBlue=  0.00f;                            
    break;
  case glvBlueColor: 
    currentColor= glvBlueColor; 
    glvRed=   0.24f; 
    glvGreen= 0.47f; 
    glvBlue=  0.60f;                            
    break;
  case glvMyOwnColor: 
    currentColor= glvMyOwnColor; 
    glvRed=   glvOwnColor[0]; 
    glvGreen= glvOwnColor[1]; 
    glvBlue=  glvOwnColor[2]; 
    glvAlpha= glvOwnColor[3];                   
    break;
  }
}


/** Return and id of the current clear color.
 *  The id is: glvBlackColor, glvWhiteColor, glvBlueColor or glvMyOwnColor
 */
int  glvViewer::glvGetClearColor (void) {return currentColor;}



/*----------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
                            INPUT / OUTPUT FUNCTIONS
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
----------------------------------------------------------------------------*/




/** Save an image ppm file of the current view.
 */
void glvViewer::glvSaveImage (void) {
  
  int viewport[4];
  glGetIntegerv (GL_VIEWPORT, viewport);
  
  int ox=  viewport[0];
  int oy=  viewport[1];
  int width=  viewport[2];
  int height= viewport[3];
  
  int n= width* height* 3;
  GLubyte *pixels = (GLubyte *) calloc( sizeof (GLubyte), n);

  if (pixels== NULL) {
    cout << "I have not enough memory to save the required image." << endl;
    cout << "Try another size." << endl;
  }
  else {
    glPixelStorei(GL_PACK_ALIGNMENT, 1); // this is a very jodida instruction
    glReadPixels (ox, oy, width, height, GL_RGB, GL_UNSIGNED_BYTE, pixels);
    FILE *file= fopen ("image.ppm", "wb");
    
    if (file== NULL) cout << "I could not open a file for writing" << endl;
    else {
      fprintf( file, "P6\n" );
      fprintf( file, "%d %d\n255\n", width, height);
      // fwrite( pixels, sizeof(GLubyte), n, file);
      for (int i= height-1; i>= 0; i--) {
	fwrite (pixels+ (i* width* sizeof(GLubyte)*3), sizeof(GLubyte), 
		width*3, file);
      }
      fclose( file );
      cout << "File image.ppm saved: " << width <<"x" << height << endl;
    }
    free (pixels);
  }
}


/** Save current options in the glv_options file.
 */
void glvViewer::glvSaveOptions (void) {

  ofstream file("glv_options");

  if (file) {
    GLint state, state1;
    
    file << "#### Rasterization" << endl;
    
    file << "window_size "  << glutGet(GLUT_WINDOW_WIDTH) << " " << glutGet(GLUT_WINDOW_HEIGHT) << endl;
    file << "window_pos " <<  glutGet(GLUT_WINDOW_X) << " " <<  glutGet(GLUT_WINDOW_Y) << endl;
    file << "game_mode " << glvGameMode << endl;
    
    file << "clear_color ";
    switch (currentColor) {
    case glvWhiteColor: file << "white" << endl; break;
    case glvBlackColor: file << "black" << endl; break;
    case glvBlueColor:  file << "blue"  << endl; break;
    case glvMyOwnColor: file << "own"   << endl; break;
    }
    file << "own_clear_color "   << glvOwnColor[0] << " " << glvOwnColor[1] << " "  
	 << glvOwnColor[2] << " " << glvOwnColor[3] << endl;
    
    glGetIntegerv(GL_SHADE_MODEL, &state);
    file << "shade_model ";
    if (state== GL_SMOOTH) file << "smooth" << endl; 
    else file << "flat" << endl;
    
    glGetIntegerv(GL_POLYGON_MODE, &state1);
    file << "polygon_mode ";
    if (state1== GL_FILL) file << "fill" << endl; 
    else file << "line" << endl;
    
    glGetIntegerv(GL_FRONT_FACE, &state);
    file << "front_face ";
    if (state== GL_CCW) file << "ccw" << endl; 
    else file << "cw" << endl;
    
    glGetIntegerv(GL_CULL_FACE_MODE, &state);
    file << "cull_face_mode ";
    if (state== GL_FRONT) file << "front" << endl; 
    else file << "back" << endl;

    file << "cull_face ";
    if (glIsEnabled(GL_CULL_FACE)) file << "yes"<< endl;
    else file << "no"<< endl;
    
    file << "#### Transformation" << endl;
    
    file << "projection ";
    if (glvProjection== glvParallel) file << "parallel"<< endl;
    else file << "perspective"<< endl;
    
    file << "camera_fovy " << glvFovy << endl;
    
    file << "normalize ";
    if (glIsEnabled(GL_NORMALIZE)) file << "yes"<< endl;
    else file << "no"<< endl;
    
    GLfloat v[4];
    GLboolean b;
    
    file << "#### Lighting" << endl;
    
    file << "Enable_Light0\t\t\t";
    if (glIsEnabled(GL_LIGHT0)) file << "yes"<< endl;
    else file << "no"<< endl;
    
    file << "Enable_Lighting\t\t\t";
    if (glIsEnabled(GL_LIGHTING)) file << "yes"<< endl;
    else file << "no"<< endl;
    
    file << "Enable_Color_Material\t\t";
    if (glIsEnabled(GL_COLOR_MATERIAL)) file << "yes"<< endl;
    else file << "no"<< endl;
    
    
    glGetBooleanv(GL_LIGHT_MODEL_TWO_SIDE, &b);
    file << "Enable_Light_Two_Side\t\t";
    if (b== GL_TRUE) file << "yes" << endl; 
    else file << "no" << endl;

    glGetBooleanv(GL_LIGHT_MODEL_LOCAL_VIEWER, &b);
    file << "Enable_Light_Local_Viewer\t";
    if (b== GL_TRUE) file << "yes" << endl; 
    else file << "no" << endl;
    
    glGetFloatv (GL_LIGHT_MODEL_AMBIENT, v);
    file << "Light_Model_Ambient\t\t" << v[0] << " " << v[1] << " " << v[2] << " " << v[3] << endl;
    
    glGetLightfv (GL_LIGHT0, GL_POSITION, v);
    file << "Light0_Position\t\t\t" << v[0] << " " << v[1] << " " << v[2] << " " << v[3] << endl;
    
    glGetLightfv (GL_LIGHT0, GL_DIFFUSE, v);
    file << "Light0_Diffuse\t\t\t" << v[0] << " " << v[1] << " " << v[2] << " " << v[3] << endl;
    
    glGetLightfv (GL_LIGHT0, GL_SPECULAR, v);
    file << "Light0_Specular\t\t\t" << v[0] << " " << v[1] << " " << v[2] << " " << v[3] << endl;
    
    glGetLightfv (GL_LIGHT0, GL_AMBIENT, v);
    file << "Light0_Ambient\t\t\t" << v[0] << " " << v[1] << " " << v[2] << " " << v[3] << endl;
    
    glGetMaterialfv (GL_FRONT, GL_AMBIENT, v);
    file << "Material_Ambient\t\t" << v[0] << " " << v[1] << " " << v[2] << " " << v[3] << endl;
    
    glGetMaterialfv (GL_FRONT, GL_DIFFUSE, v);
    file << "Material_Diffuse\t\t" << v[0] << " " << v[1] << " " << v[2] << " " << v[3] << endl;
    
    glGetMaterialfv (GL_FRONT, GL_SPECULAR, v);
    file << "Material_Specular\t\t" << v[0] << " " << v[1] << " " << v[2] << " " << v[3] << endl;
    
    glGetMaterialfv (GL_FRONT, GL_EMISSION, v);
    file << "Material_Emission\t\t" << v[0] << " " << v[1] << " " << v[2] << " " << v[3] << endl;
    
    glGetMaterialfv (GL_FRONT, GL_SHININESS, v);
    file << "Material_Shininess\t\t" << v[0] << endl;
    //file.flush();
    //file.close();
  }
}


/** Reload the glv_options file and set again the rendering values.
 */
void glvViewer::glvReloadOptions (void) {

  ifstream in ("glv_options");
  if (in) {
    glvRead(in);
  }
  glvRenderOptions();
}


/** Read the glv_options file.
 */
void glvViewer::glvRead (ifstream &in) {

  char option[60], value[30];

  in >> option >> value; // read the ### Rasterization
  
  in >> option >> glvViewerWindowWidth >> glvViewerWindowHeight;
  in >> option >> glvViewerWindowPosX  >> glvViewerWindowPosY;
  in >> option >> glvGameMode;
  
  in >> option >> value;
  
  switch (value[2]) {
  case 'i': currentColor= glvWhiteColor; 
    glvRed= 1.00f; glvGreen= 1.00f; glvBlue= 1.00f;break;
  case 'a': currentColor= glvBlackColor; 
    glvRed= 0.00f; glvGreen= 0.00f; glvBlue= 0.00f; break;
  case 'u': currentColor= glvBlueColor; 
    glvRed= 0.24f; glvGreen= 0.47f; glvBlue= 0.60f; break;
  case 'n': currentColor= glvMyOwnColor; break;
  }
  in >> option >> 
    glvOwnColor[0] >> glvOwnColor[1] >> glvOwnColor[2] >> glvOwnColor[3];
  if (currentColor== glvMyOwnColor) { 
    glvRed= glvOwnColor[0]; 
    glvGreen= glvOwnColor[1]; 
    glvBlue= glvOwnColor[2]; 
    glvAlpha= glvOwnColor[3];
  }
  
  in >> option >> value;
  if (value[0]=='s') glvShadeModel= GL_SMOOTH;
  else glvShadeModel= GL_FLAT;
  
  in >> option >> value;
  if (value[0]=='f') glvPolygonMode= GL_FILL;
  else glvPolygonMode= GL_LINE;
  
  in >> option >> value;
  if (value[1]=='c') glvFrontFace= GL_CCW;
  else glvFrontFace= GL_CW;
  
  in >> option >> value;
  if (value[0]=='f') glvCullFace= GL_FRONT;
  else glvCullFace= GL_BACK;
  
  in >> option >> value;
  if (value[0]=='y') glvEnableCullFace= GL_TRUE;
  else glvEnableCullFace= GL_FALSE;
  
  
  in >> option >> value; // read the ### Transformation
  
  in >> option >> value;
  if (value[1]=='e') glvProjection= glvPerspective;
  else glvProjection= glvParallel;
  
  in >> option >> glvFovy;
  
  in >> option >> value;
  if (value[0]=='y') glvNormalize= GL_TRUE;
  else glvNormalize= GL_FALSE;

  
  
  in >> option >> value; // read the ### Lighting
  
  in >> option >> value;
  if (value[0]=='y') glvEnableLight0= GL_TRUE;
  else glvEnableLight0= GL_FALSE;
  
  in >> option >> value;
  if (value[0]=='y') glvEnableLighting= GL_TRUE;
  else glvEnableLighting= GL_FALSE;
  
  in >> option >> value;
  if (value[0]=='y') glvEnableColorMaterial= GL_TRUE;
  else glvEnableColorMaterial= GL_FALSE;
  
  in >> option >> value;
  if (value[0]=='y') glvLightTwoSide= GL_TRUE;
  else glvLightTwoSide= GL_FALSE;
  
  in >> option >> value;
  if (value[0]=='y') glvLightLocalViewer= GL_TRUE;
  else glvLightLocalViewer= GL_FALSE;
  
  in >> option >> glvModelAmbient[0] >>   glvModelAmbient[1] >>   
    glvModelAmbient[2] >>   glvModelAmbient[3];
  in >> option >> glvLight0Position[0] >> glvLight0Position[1] >> 
    glvLight0Position[2] >> glvLight0Position[3];
  in >> option >> glvLight0Diffuse[0] >>  glvLight0Diffuse[1] >>  
glvLight0Diffuse[2] >>  glvLight0Diffuse[3];
  in >> option >> glvLight0Specular[0] >> glvLight0Specular[1] >> 
glvLight0Specular[2] >> glvLight0Specular[3];
  in >> option >> glvLight0Ambient[0] >>  glvLight0Ambient[1] >>  
glvLight0Ambient[2] >>  glvLight0Ambient[3];
  
  in >> option >> glvMaterialAmbient[0] >>  glvMaterialAmbient[1] >>  
glvMaterialAmbient[2] >>  glvMaterialAmbient[3];
  in >> option >> glvMaterialDiffuse[0] >>  glvMaterialDiffuse[1] >>  
glvMaterialDiffuse[2] >>  glvMaterialDiffuse[3];
  in >> option >> glvMaterialSpecular[0] >> glvMaterialSpecular[1] >> 
glvMaterialSpecular[2] >> glvMaterialSpecular[3];
  in >> option >> glvMaterialEmission[0] >> glvMaterialEmission[1] >> 
glvMaterialEmission[2] >> glvMaterialEmission[3];
  in >> option >> glvMaterialShininess[0];
}

